% SPARSE MATRICES - from the "Using Matlab" guide - Ch. 9
% prepared by Gilberto E. Urroz, August 2004
%
echo on
% Sparse matrices: contain significant number of zero elements
% Only need to store those non-zero elements
pause
% MATLAB uses 3 arrays to store sparse matrices.
% For a m x n sparse matrix with nnz nonzero entries:
% * First array (of length nnz) contains all nonzero elements
% * Second array (of length nnz) contains integer row indices for nonzero elements
% * Third array (of length n) contains pointers to start of each column
% Required storage: nnz floating-point numbers and nnz+n integers
%
pause
% Convert full matrix to sparse matrix  -- use function 'sparse'
A = [0  0 0 5; 0 2 0 0; 1 3 0 0; 0 0 4 0 ]
pause

S = sparse(A)
pause

B = full(S)
pause

% Direct entry of sparse matrix: sparse([rows],[cols],[nonzeros],m,n)
S = sparse([3 2 3 4 1], [1 2 2 3 4], [1 2 3 4 5], 4, 4)
pause

% Other examples of sparse matrices:
n=5; D = sparse(1:n,1:n,-2*ones(1,n),n,n)
pause

E = sparse(2:n,1:n-1,ones(1,n-1),n,n)
pause

S = E + D + E'
pause

F = full(S)
pause

% NOTE: previous result shows the numerical set up for a second order difference equations
% (a numerical method used for solving second order ODE or PDE)
pause
 
% Creating sparse matrix out of diagonal elements, S = spdiags(B,d,m,n):
% The matrix to be created is S (mxn) with nonzero elements in p diagonals.
% Matrix B has size:  min(m,n) x p, and its columns will become diagonals of S.
% d is a vector of length p with integer elements referring to the diagonals
% of S to be populated.   
% Example:
B = [41 11 0; 52 22 0; 63 33 13; 74 44 24]
pause

d = [-3; 0; 2]
pause

% We'll use B and d to create a matrix S with 7 rows and 4 columns:
S = spdiags(B,d,7,4)
pause

% To see the full matrix use:
full(S)
pause

% compare to B
B
pause

% Use command 'whos' to see details on variables
whos
pause

% Information about nonzero elements
%   nnz(S) --> number of nonzero elements in sparse matrix S
%   nonzeros(S) --> vector with all nonzero elements in S
%   nzmax(S) --> storage space for nonzero elements in S
%
% Examples:
nnz(S)
pause

nonzeros(S)
pause

nzmax(S)
pause

% Viewing sparse matrix graphically: use function 'spy'
spy(S)
pause

% The following example uses a sparse matrix called 'west0479' available
% in your Matlab installation:
load west0479
pause

nnz(west0479)
pause

spy(west0479)
pause

% See pages 9-15 and 9-16 for explanation of the following exercise.
% "Graphing Using Adjacency Matrices"
A = [0 1 0 1; 1 0 1 0; 0 1 0 1; 1 0 1 0]
pause

xy = [1 3; 2 1; 3 3; 2 5]
pause

gplot(A,xy)
pause

help gplot
pause

% See pages 9-16 through 9-19 in the "Using Matlab" guide for the
% explanation of the following exercise: The Bucky Ball
%
% Bucky ball adjacency matrix:
B = bucky;
pause

% Alternatively, use: [B,v] = bucky; (this includes coordinates)
[B,v] = bucky
pause

gplot(B,v)
pause

% Use the following commands to show only half of the nodes:
k = 1:30;
gplot(B(k,k),v);
axis square
for j = 1:30, text(v(j,1),v(j,2), num2str(j)); end;
spy(B)
pause

% Graphics of powers of B
figure(1); spy(B^2)
figure(2); spy(B^3)
figure(3); spy(B^4)
figure(4); spy(B^8)
pause

% Showing four graphs is a single figure
figure(5);n = [2, 3, 4, 8];
for k = 1:4
         subplot(2,2,k);spy(B^n(k));
end
pause

% Sparse matrix operations
% Read pp. 9-23 and 9-24 for information on the subjects of: 
% * Computational Considerations
% * Standard Mathematical Operations
%
% Permutation and reordering
% ==========================
% Two representations of permutation:
% 1 - Using a permutation matrix P which will act on:
%     a. Rows of S as P*S
%     b. Columns of S as S*P'
%
% 2 - Using a permutation vector p containing a permutation of
%     1:n acting on:
%     a. Rows of S as S(p,:)
%     b. Columns of S as S(:,p)
%
% Example:
p = [1 3 4 2 5]            % a permutation of 1,2,3,4,5
pause

I = eye(5,5)               % identity matrix 5 x 5 (a sparse matrix indeed)
pause

P = I(p,:)                 % permutation of I on rows
pause

% Creating a new sparse matrix using diagonals             
e = ones(4,1)
pause

S = diag(11:11:55) + diag(e,1) + diag(e,-1)
pause

% Permutations on S using permutation matrix P
% On rows:
P*S
pause

% On columns
S*P'
pause

P        % compare with P
pause

% Permutations on S using permutation vector p
S(p,:)
pause

S(:,p)
pause

% Note: to create a sparse identity matrix (it will store only the diagonal)
%       use function 'speye(n)', example:
Is = speye(5)
pause

% Repeating this exercise with Is:
P = Is(p,:)
pause

p = (1:5)*P' 
pause

% Note: now both P and p are sparse matrices too
S
pause

% Convert S into sparse:
S = sparse(S)
pause

% permutations using P:
P*S
pause

% permutations using p
S
pause

S(p,:)
pause
% Read pp. 9-26 through 9-32 in the "Using Matlab" guide for the concepts
% of reordering sparse matrices, and using sparse matrices for matrix
% factorization (LU, QR, and Cholesky factorization)
pause

% Read pp. 9-33 through 9-35 to learn about the application of sparse matrices
% to the solution of systems of linear equations.
pause

% Read pp. 9-36 through 9-38 in the "Using Matlab" guide for applications of 
% certain functions to the calculation of eigenvalues in sparse matrices.

% End of Script

